import random
import algorithm as a

# path to russian and english vocabulary bank (1000 words each)
pathRussian = "russian.txt"
pathEnglish = "english.txt"


def deleteEqualValues(words):
    deletedEqualValues = {}
    for comb in words:
        areUnique = True
        for item in words[comb]:
            for initialList in [item for item in list(deletedEqualValues.values())]:
                for initialItem in initialList:
                    if item == initialItem:
                        areUnique = False
        if areUnique:
            deletedEqualValues[comb] = words[comb]
    return deletedEqualValues


def bubbleSort(items):
    for i in range(len(items)):
        for j in range(len(items) - 1 - i):
            if len(items[j]) < len(items[j + 1]):
                items[j], items[j + 1] = items[j + 1], items[j]
    return items


def sortWordsInRelevantWay(words):
    # sorting dict by length of the comb
    listOfKeys = [item for item in words]
    listOfKeys = bubbleSort(listOfKeys)
    # having a sorted list instead of dict
    wordsList = [{item: words[item]} for item in listOfKeys]

    # splitting list in the middle into two parts
    # and getting a new list sorted in a relevant way
    sortedWordsList = []
    num = 1
    middle = len(wordsList) // 2
    for index in range(middle - 1):
        for k, v in wordsList[middle + num - 1].items():
            sortedWordsList.append({k: v})
        for k, v in wordsList[middle - num].items():
            sortedWordsList.append({k: v})
        num += 1
    return sortedWordsList


def selectCombinations(words):
    # all the possible letters combinations in all the words
    combinations = [word[start:end] for word in words for start in range(len(word) + 1)
                    for end in range(len(word) + 1) if len(word[start:end]) != 0]

    wordsDict = {}  # {comb: [allOfTheWordsThatContainSuchComb]}
    for word in words:
        for comb in combinations:
            if word.find(comb) != -1:  # if comb in the word
                if comb in wordsDict:  # if combination already in the the dict
                    if word not in wordsDict[comb]:  # if particular word not in the list
                        wordsDict[comb].append(word)
                else:
                    wordsDict[comb] = [word]
    # delete words which have no matches
    return {k: v for k, v in wordsDict.items() if len(v) > 1}


def getWords(path):
    with open(path, "r", encoding="windows-1251") as f:
        allWords = [word.replace("\n", "") for word in f]
        allWords = [word for word in allWords if word.find("-") == -1]
        random.shuffle(allWords)
        return allWords[:100]  # shuffling all of the words and getting 100 of them


def fillEmptySpaces(matrix, lang):
    russian = ['а', 'б', 'в', 'г', 'д', 'е', 'ж', 'з', 'и', 'й', 'к', 'л', 'м', 'н', 'о', 'п',
               'р', 'с', 'т', 'у', 'ф', 'х', 'ц', 'ч', 'ш', 'щ', 'ъ', 'ы', 'ь', 'э', 'ю', 'я']
    english = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'g', 'k', 'l', 'm',
               'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z']
    letters = russian if lang == pathRussian else english
    for rowIndex in range(len(matrix)):
        for columnIndex in range(len(matrix[rowIndex])):
            if matrix[rowIndex][columnIndex] is None:
                matrix[rowIndex][columnIndex] = letters[random.randint(0, 9)]
    return matrix


def getMatrix(lang):
    lang = pathRussian if lang == "ru" else pathEnglish
    words = getWords(lang)  # getting all words

    # selecting words combinations {partOfTheWord: [allOfTheWordsThatContainSuchPart]}
    words = selectCombinations(words)
    # deleting equal values (some comb could refer to the same value)
    words = deleteEqualValues(words)
    # sorting words by length of comb and getting a new list starting from the middle
    words = sortWordsInRelevantWay(words)

    # getting only first half of the list and shuffle it
    random.shuffle(words[:len(words) // 2])

    # getting a matrix and [allWordsInTheMatrix] from Algorithm file
    matrix, words = a.getMatrixAndWords(words)
    matrix = fillEmptySpaces(matrix, lang)
    return matrix, words



import random
isDone = False
matrix = [[None for j in range(5)] for i in range(5)]


def neighbourFunction(i, j, x, y):
    # distance between two points formula
    return (j ** 2 - 2 * j * x + x ** 2) + (i ** 2 - 2 * i * y + y ** 2)


def getFreeNeighbours(i, j):
    global matrix
    # getting all the accessible cells using neighbour function
    # to check whether cell accessible or not
    accessibleCells = [[y, x] for y in range(5) for x in range(5)
                       if neighbourFunction(i, j, x, y) == 1
                       or neighbourFunction(i, j, x, y) == 2]

    # returning free cells only (if they are equal to None)
    freeNeighbours = [pair for pair in accessibleCells if matrix[pair[0]][pair[1]] is None]
    random.shuffle(freeNeighbours)
    return freeNeighbours


def getStackFreeInitialCells():
    global matrix
    # stack of free cells at the beginning
    stackFreeInitialCells = [[i, j] for j in range(len(matrix)) for i in range(len(matrix[j])) if matrix[i][j] is None]
    random.shuffle(stackFreeInitialCells)
    return stackFreeInitialCells


def fillWord(stackLetters, stackWay, stackFreeCells):
    global isDone  # global variable True if the word has been successfully inserted

    # Step 1 - pick i and j to insert the letter
    # getting the first item from free cells from the end
    stackFreeCellsIJ = stackFreeCells[-1].pop(0) \
        if stackFreeCells[-1] and len(stackFreeCells[-1]) != 0 else [None, None]
    i, j = stackFreeCellsIJ[0], stackFreeCellsIJ[1]

    # Step 2 - insert the letter
    # or delete the previous letter if there are no free accessible cells
    if i is not None and j is not None:
        matrix[i][j] = stackLetters.pop(0)  # inserting to the matrix
        stackWay.append([i, j])  # adding coordinates to the stack way
        if len(stackLetters) != 0:
            stackFreeCells.append(getFreeNeighbours(i, j))  # adding free neighbours to the stack of free cells
            fillWord(stackLetters, stackWay, stackFreeCells)
            return stackWay if isDone else None
        else:
            # once stackLetters is empty (we have successfully inserted the word) we mark isDone variable as True
            isDone = True
            return stackWay if isDone else None
    else:
        for place in reversed(stackWay):
            deletedLetter = matrix[place[0]].pop(place[1])
            matrix[place[0]].insert(place[1], None)
            stackLetters.insert(0, deletedLetter)  # adding to the stack of letters the last letter
            # deleting the last element on the way and of free cells (it is always an empty list)
            stackWay.pop(), stackFreeCells.pop()
            fillWord(stackLetters, stackWay, stackFreeCells)
            return stackWay if isDone else None
        return False


def getMatrixAndWords(words):
    global matrix

    combinations = [comb for item in words for comb in item]  # all the combinations
    words = [item[comb] for item in words for comb in item]  # all the words with such combination

    filledWords = {}  # {word in the matrix: [coordinates]}

    for index in range(len(combinations)):
        stackWay = []  # stack with coordinates of way filling the word
        stackFreeCells = [getStackFreeInitialCells()]  # stack with free cells in each step of insertion
        way = fillWord([letter for letter in combinations[index]], stackWay, stackFreeCells)
        if way:
            comb = combinations[index]
            for word in words[index]:
                i1, j1 = way[0][0], way[0][1]  # first coordinate
                i2, j2 = way[len(comb) - 1][0], way[len(comb) - 1][1]  # last coordinate
                # getting indexes of starting and ending comb
                endIndexFirstLetter = word.find(comb)
                startIndexLastLetter = endIndexFirstLetter + len(comb)

                # getting the first and the second part of the word
                # (the first should be reversed as we are going to insert the word backwards)
                firstPartStackLetters = list(reversed([word[index] for index in range(endIndexFirstLetter)]))
                a = fillWord(firstPartStackLetters, stackWay=[], stackFreeCells=[getFreeNeighbours(i1, j1)]) \
                    if len(firstPartStackLetters) != 0 else True

                secondPartStackLetters = [word[index] for index in range(startIndexLastLetter, len(word))]
                b = fillWord(secondPartStackLetters, stackWay=[], stackFreeCells=[getFreeNeighbours(i2, j2)]) \
                    if len(secondPartStackLetters) != 0 else True
                if not a and b and len(getStackFreeInitialCells()) < 5:
                    return matrix, filledWords
                elif a and b:
                    a = list(reversed(a)) if not isinstance(a, bool) else []
                    b = b if not isinstance(b, bool) else []
                    filledWords[word] = a + way + b  # adding word and path to the dict
        else:
            return matrix, filledWords
    return matrix, filledWords



import main as m

def getMatrix(lang="en"):
	return (m.getMatrix(lang))


def printMatrix(matrix):
    for row in matrix:
        cellStr = ""
        for cell in row:
            cellStr += cell + ' '
        print(cellStr)


def printWords(words):
    for word in words:
        print(word + " - " + str(words[word]))